from __future__ import absolute_import
import pudgy

from snorkel.views import ViewBase
from snorkel.components import *

from .time import TimeView
import os

class AreaView(TimeView):
    NAME="area"
    BASE="time"
    DISPLAY_NAME="Stacked Area View"
    BASE_DIR=os.path.dirname(__file__)



import pudgy
import flask
import os
import time

import addict

from flask import redirect, url_for
from flask_security import core, current_user

from . import auth, components, results, admin, config, grafana
from . import fastjson as json

from .pages import QueryPage, DatasetsPage, HomePage, UserPage
from .auth import needs_login
from .backend.sybil import SybilBackend
from .util import return_json




app = flask.Flask(__name__)

from werkzeug.contrib.fixers import ProxyFix
app.wsgi_app = ProxyFix(app.wsgi_app)


pudgy.register_blueprint(app)

app.config.update({
    "SESSION_COOKIE_NAME" : "slite"
})

@app.before_request
def start_logger():
    flask.request._sample = addict.Dict()
    flask.request._sample.time = time.time()

@app.after_request
def send_logs(req):
    if not flask.request._sample:
        return

    url = flask.request.url
    route = 'unknown'
    if flask.request.url_rule:
        route = flask.request.url_rule.rule
    duration = time.time() - flask.request._sample.time
    flask.request._sample.duration = duration
    flask.request._sample.route = route
    flask.request._sample.url = url
    ua = flask.request.user_agent

    flask.request._sample.user_agent = {
        "browser" : ua.browser,
        "platform" : ua.platform,
        "version" : ua.version,
        "language" : ua.language,
        "ip" : flask.request.remote_addr
    }

    bs = SybilBackend()
    try:
        bs.ingest("slite@pagestats", [flask.request._sample])
    except:
        pass

    return req


@app.route('/')
def get_index():
    if current_user.is_authenticated:
        return redirect(url_for('get_datasets'))

    if not config.SHOW_TOUR:
        return redirect(url_for('get_datasets'))

    return HomePage(template="welcome.html").render()

@app.route('/tour')
def get_tour():
    carousel = components.TourCarousel()
    return HomePage(template="tour.html", carousel=carousel).render()

@app.route('/pkg/status')
def get_status():
    return "OK", 200

@app.route('/user')
@needs_login
def get_user_settings():
    return UserPage(template='user_settings.html', auth_token=current_user.get_auth_token()).render()

@app.route('/datasets')
@needs_login
def get_datasets():
    return DatasetsPage(template="datasets.html").render()

@app.route('/query')
@needs_login
def get_view():
    query = flask.request.args

    query_id = query.get("h")
    table = query.get('table')
    view = query.get('view', 'table')
    sq = None
    if query_id:
        sq = results.get_by_hashid(query_id)
        if sq:
            sq = sq.pop()
            table = sq["table"]
            view = sq["parsed"]["view"]

    return QueryPage(template="query.html", table=table, view=view, saved=sq).pipeline()

@app.route('/query/grafana')
def get_grafana():
    query = flask.request.args
    return grafana.run_query(query)

# this route expects JSON data
# table = the table to import into
# samples = the samples to import
@app.route('/data/import', methods=['POST'])
@needs_login
def post_data():
    bs = SybilBackend()

    # data can either be sent as form data or as JSON data
    # so we first try reading the JSON
    args = flask.request.json
    if not args:
        args = flask.request.form

    table = args.get("table", None)

    if not table:
        dataset = args.get("dataset")
        subset = args.get("subset")
        if dataset and subset:
            table = '%s@%s' % (dataset, subset)

    if not table:
        return return_json({'error' : "No table specified"})

    samples = args.get("samples", None)
    if isinstance(samples, (str, unicode)):
        try:
            samples = json.loads(samples)
        except:
            samples = None

    if not samples:
        return return_json({'error' : "No samples specified"})

    bs.ingest(table, samples)

    return return_json({'success':True, "num_samples" : len(samples)})


admin.install(app)
auth.install(app)
components.install(app)

if __name__ == "__main__":
    app.run(port=2333, use_reloader=False)




import pudgy

from snorkel.views import ViewBase
from snorkel.components import *

from ..snorkel_basic_views import TableView
import os

class DrilldownView(TableView):
    NAME="drilldown"
    BASE="table"
    DISPLAY_NAME="Drilldown View"
    BASE_DIR=os.path.dirname(__file__)



#!/usr/bin/env python
from __future__ import print_function

import os
import pipes
import tempfile
import shutil
import sys
import random

KEEP_RESULTS=False
BASEDIR=None

# from https://stackoverflow.com/questions/5574702/how-to-print-to-stderr-in-python
def debug(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def debug_file(filename, title):
    return print_file(filename, title, fd=sys.stderr)

def write_to_file(filename, data):
    if isinstance(data, bytes):
        data = data.decode("utf-8")

    with open(filename, "w") as f:
        f.write(data or "")

def run_subprocess(cmd, stdin=None, wait=True):
    import subprocess
    import shlex

    # if we are pushing cmd through the shell, we quote it
    process = subprocess.Popen(shlex.split(cmd), stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.PIPE)

    if wait:
        if isinstance(stdin, str):
            stdin = stdin.encode("utf-8")

        stdout,stderr = process.communicate(stdin)
        return stdout.strip(), stderr
    else:
        return "", ""

def cleanup():
    if KEEP_RESULTS:
        return

    if BASEDIR:
        shutil.rmtree(BASEDIR)

def check_connection(host):
    full_cmd = "ssh -O check %s " % (host)
    out, err = run_subprocess(full_cmd)
    err = err.decode("utf-8")

    if err.find("running") != -1:
        return

    run_subprocess("ssh -O exit %s " % (host))
    debug("NO MASTER RUNNING FOR HOST %s, STARTING MASTER" % host)
    run_subprocess("ssh -C -N -n %s #master process" % host, wait=False)

    import time
    # when we check our connection, we'll wait up to a second while
    # establishing it
    incr = 0.1
    steps = int(1 / incr)
    for _ in range(steps):
        time.sleep(incr)
        out, err = run_subprocess(full_cmd)
        err = err.decode("utf-8")
        if err.find("running") != -1:
            break


def setup_dirs():
    global BASEDIR, RESULT_DIR, OUTPUT_DIR
    global COMBINE_LOG, COMBINE_RESULTS

    BASEDIR = tempfile.mkdtemp()
    RESULT_DIR = os.path.join(BASEDIR, "results")
    OUTPUT_DIR = os.path.join(BASEDIR, "output")

    COMBINE_LOG = os.path.join(BASEDIR, "master.log")
    COMBINE_RESULTS = os.path.join(BASEDIR, "master.results")

    os.makedirs(RESULT_DIR)
    os.makedirs(OUTPUT_DIR)

    debug("collecting results in", BASEDIR)

def ingest_on_host(host, working_dir, bin, table, samples):
    check_connection(host)

    import pipes

    table = pipes.quote(table)

    full_cmd = "ssh -C %s \"cd \"%s\" && \"%s\" ingest --debug -table %s\"" % (host, working_dir, bin, table)
    out, err = run_subprocess(full_cmd, stdin=samples)
    write_to_file(os.path.join(RESULT_DIR, "%s.results" % host), out)
    write_to_file(os.path.join(OUTPUT_DIR, "%s.log" % host), err)
    print("DEBUG", err)

    debug("%s finished" % host)
    debug("Ingested %s samples" % len(samples.split('\n')))

def read_host_info():
    global HOSTINFO
    global HOST_FILE
    host_file = os.environ.get("MSYBIL")
    with open(host_file) as f:
        hostinfo = f.readlines()

    hosts = []
    for line in hostinfo:
        line = line.strip()
        if not line:
            continue

        if line[0] == '#':
            continue
        hosts.append(line.split(" "))


    HOSTINFO = hosts
    debug("host info", hosts)

def read_samples():
    return sys.stdin.read()

def _main():
    global CMD
    setup_dirs()
    table = sys.argv[1]
    debug("command to run is: sybil %s"  % table)

    read_host_info()

    samples = read_samples()
    bin="sybil"
    working_dir="~"
    line = random.choice(HOSTINFO)
    host = line[0]

    check_connection(host)

    if len(line) > 1:
        working_dir = line[1]
    if len(line) > 2:
        bin = line[2]

    ingest_on_host(host, working_dir, bin, table, samples)

if "DEBUG" in os.environ:
    KEEP_RESULTS=True

SYBIL_BIN = os.path.join(os.path.dirname(__file__), "bin", "sybil")
def main():
    global SYBIL_BIN
    if not os.path.exists(SYBIL_BIN):
        SYBIL_BIN, _ = run_subprocess("which sybil")

    try:
        _main()
    finally:
        cleanup()

if __name__ == "__main__":
    main()

# vim: set syntax=python



#!/usr/bin/env python
from __future__ import print_function

import os
import pipes
import tempfile
import shutil
import sys
import threading

KEEP_RESULTS=False
BASEDIR=None

# from https://stackoverflow.com/questions/5574702/how-to-print-to-stderr-in-python
def debug(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def run_subprocess(cmd, stdin=None, wait=True):
    import subprocess
    import shlex

    # if we are pushing cmd through the shell, we quote it
    process = subprocess.Popen(shlex.split(cmd), stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.PIPE)
    if wait:
        stdout,stderr = process.communicate(stdin)
        return stdout.strip(), stderr
    else:
        return "", ""

def cleanup():
    if KEEP_RESULTS:
        return

    if BASEDIR:
        shutil.rmtree(BASEDIR)

def setup_dirs():
    global BASEDIR, RESULT_DIR, OUTPUT_DIR
    global COMBINE_LOG, COMBINE_RESULTS

    BASEDIR = tempfile.mkdtemp()
    RESULT_DIR = os.path.join(BASEDIR, "results")
    OUTPUT_DIR = os.path.join(BASEDIR, "output")

    COMBINE_LOG = os.path.join(BASEDIR, "master.log")
    COMBINE_RESULTS = os.path.join(BASEDIR, "master.results")

    os.makedirs(RESULT_DIR)
    os.makedirs(OUTPUT_DIR)

    debug("collecting results in", BASEDIR)


def print_file(filename, title="", fd=sys.stdout, force=False):
    filetype,err = run_subprocess("file '%s'" % filename)
    filetype = filetype.lower()

    filetype = filetype.decode("utf-8")

    if filetype.find("ascii") != -1 or filetype.find("unicode") != -1 or force:
        with open(filename) as f:
            filedata = f.read().strip()
            if title:
                debug(title)
            print(filedata, file=fd)

def debug_file(filename, title):
    return print_file(filename, title, fd=sys.stderr)

def write_to_file(filename, data):
    with open(filename, "wb") as f:
        f.write(data or b"")

def run_local_commands(cmd):
    out,err = run_subprocess("%s %s" % (SYBIL_BIN, cmd))
    print(out)


def check_connection(host):
    full_cmd = "ssh -O check %s " % (host)
    out, err = run_subprocess(full_cmd)
    err = err.decode("utf-8")
    if err.find("running") != -1:
        return

    run_subprocess("ssh -O exit %s " % (host))
    debug("NO MASTER RUNNING FOR HOST %s, STARTING MASTER" % host)
    run_subprocess("ssh -C -N -n %s #master process" % host, wait=False)

    import time
    # when we check our connection, we'll wait up to a second while
    # establishing it
    incr = 0.1
    steps = int(1 / incr)
    for _ in range(steps):
        time.sleep(incr)
        out, err = run_subprocess(full_cmd)
        err = err.decode("utf-8")
        if err.find("running") != -1:
            break


def run_command_on_host(host, working_dir, bin):
    check_connection(host)

    query_flag_file = os.path.join(BASEDIR, "query_flags.gob")
    with open(query_flag_file, "rb") as f:
        query_flag_data = f.read()

    full_cmd = "ssh -C %s \"cd \"%s\" && \"%s\" query -decode-flags\"" % (host, working_dir, bin)
    out, err = run_subprocess(full_cmd, stdin=query_flag_data)
    write_to_file(os.path.join(RESULT_DIR, "%s.results" % host), out)
    write_to_file(os.path.join(OUTPUT_DIR, "%s.log" % host), err)

    debug("%s finished" % host)

def run_remote_commands(cmd):
    global HOST_FILE
    debug("*** running command on remote hosts")
    debug_file(HOST_FILE, "*** host info is")

    query_flag_data, err = run_subprocess("%s %s -encode-flags -encode-results" % (SYBIL_BIN, cmd))
    query_flag_file = os.path.join(BASEDIR, "query_flags.gob")

    write_to_file(query_flag_file, query_flag_data)

    threads = []
    for line in HOSTINFO:
        bin="sybil"
        working_dir="~"
        host = line[0]

        if len(line) > 1:
            working_dir = line[1]
        if len(line) > 2:
            bin = line[2]



        t = threading.Thread(target=run_command_on_host, args=(host, working_dir, bin))
        threads.append(t)
        t.start()

    for t in threads:
        t.join()

def print_remote_results():
    for line in HOSTINFO:
        host = line[0]

        log_file = os.path.join(OUTPUT_DIR, "%s.log" % host)
        result_file = os.path.join(RESULT_DIR, "%s.results" % host)

        debug_file(log_file, "*** %s output" % host)
        debug_file(result_file, "*** %s output" % host)

def aggregate_remote_results():
    full_cmd = "%s aggregate \"%s\"" % (SYBIL_BIN, RESULT_DIR)
    query_flag_file = os.path.join(BASEDIR, "query_flags.gob")
    with open(query_flag_file, "rb") as f:
        query_flag_data = f.read()

    out, err = run_subprocess(full_cmd, query_flag_data)

    combine_log = os.path.join(OUTPUT_DIR, COMBINE_LOG)
    combine_results = os.path.join(RESULT_DIR, COMBINE_RESULTS)
    write_to_file(combine_results, out)
    write_to_file(combine_log, err)

    debug_file(combine_log, "*** aggregator output")
    print_file(combine_results, "***combined results", force=True)

def read_host_info():
    global HOSTINFO
    global HOST_FILE
    host_file = os.environ.get("MSYBIL")
    with open(host_file) as f:
        hostinfo = f.readlines()

    hosts = []
    for line in hostinfo:
        line = line.strip()
        if not line:
            continue

        if line[0] == '#':
            continue
        hosts.append(line.split(" "))


    HOSTINFO = hosts
    debug("host info", hosts)

    HOST_FILE = os.path.join(BASEDIR, "hosts")
    with open(HOST_FILE, "w") as f:
        for line in HOSTINFO:
            f.write(" ".join(line))
            f.write("\n")


def _main():
    global CMD
    setup_dirs()
    operator = None
    if len(sys.argv) > 1:
        operator = sys.argv[1]

    cmd = " ".join(sys.argv[1:])
    debug("command to run is: sybil %s"  % cmd)
    if operator == "query":
        read_host_info()
        run_remote_commands(cmd)
        print_remote_results()
        aggregate_remote_results()
    else:
        run_local_commands(cmd)


if "DEBUG" in os.environ:
    KEEP_RESULTS=True

SYBIL_BIN = os.path.join(os.path.dirname(__file__), "bin", "sybil")
def main():
    global SYBIL_BIN
    if not os.path.exists(SYBIL_BIN):
        SYBIL_BIN, _ = run_subprocess("which sybil")

    try:
        _main()
    finally:
        cleanup()

if __name__ == "__main__":
    main()

# vim: set syntax=python



from __future__ import print_function
import pudgy
import flask
import time

from .views import get_view_by_name
from .components import QuerySidebar, UserButton, UserModal
from .auth import rpc_login_required

from . import backend, results, presenter, rbac
from .components import UIComponent, Selector

import werkzeug
import os
from collections import defaultdict

import flask_security

from .query_spec import QuerySpec

from .util import time_to_seconds, string_dict

try:
    from urllib import unquote_plus
except:
    from urllib.parse import unquote_plus

from . import fastjson as json

class ViewArea(UIComponent, pudgy.JinjaComponent, pudgy.BackboneComponent, pudgy.ClientBridge):
    pass

@pudgy.Virtual
class Page(pudgy.FlaskPage):
    NAMESPACE = "pages"

class HomePage(Page, pudgy.SassComponent):
    pass

class UserPage(Page):
    pass

class DatasetsPage(Page, pudgy.BackboneComponent, pudgy.SassComponent):
    def __prepare__(self):
        bs = backend.SybilBackend()
        self.context.tables = filter(lambda t: rbac.check("query", t), bs.list_tables())


        self.context.tables = list(self.context.tables)
        self.context.tables.sort()
        groups = defaultdict(list)

        self.context.table_groups = groups

        groupings = []
        visited = {}
        for t in self.context.tables:
            if t in visited:
                continue

            visited[t] = 1

            tokens = t.split("@")
            if len(tokens) > 1:
                superset = tokens[0]
                dataset = " ".join(tokens[1:])
                groups[superset].append(dataset)
            else:
                groups[t].append("")


        self.context.groups = list(sorted(groups.keys()))
        self.context.user_button = UserButton()
        self.context.user_modal = UserModal()


def read_filters(query, i='query'):
    import ast

    filters = []
    try:
        filters = query.get('filters')
        if type(filters) == str:
            filters = ast.literal_eval(filters)

        filters = filters[i]
    except Exception as e:
        print("FILTER ERR", e)

    return filters


import datetime
def epoch(d):
    return (d - datetime.datetime(1970, 1, 1)).total_seconds()

class QueryPage(Page, pudgy.SassComponent, pudgy.BackboneComponent, pudgy.ServerBridge):
    def __prepare__(self):
        # locate the potential views
        query = QuerySpec(flask.request.args)
        self.context.error = None
        has_saved = False
        if "saved" in self.context and self.context.saved:
            query = QuerySpec(self.context.saved.parsed)
            has_saved = True


        table = self.context.table
        view = query.get('view', 'table')

        pr = presenter.GetPresenter(table)

        bs = backend.SybilBackend()
        bs.clear_cache(table=table)

        try:
            table_info = bs.get_table_info(table)
            tables = filter(lambda t: rbac.check("query", t), bs.list_tables())
        except Exception as e:
            self.context.error = "Couldn't read table info for table %s" % (table)
            return


        table_selector = Selector(name="table", selected=table, options=tables)

        VwClass = get_view_by_name(view)
        view = VwClass()
        view.context.update(metadata=table_info, presenter=pr, query=query)
        # its up to a view to decide on marshalling its data to client,
        # but we auto marshal the table metadata and query for every view
        view.marshal(metadata=table_info, query=query)

        viewarea = ViewArea()
        if self.context.saved:
            sq = self.context.saved

            if not sq.results:
                viewarea.call("no_results", {
                    "errmsg" : "No results found for this query"
                });
            view.context.update(sq.parsed, results=sq.results, compare=sq.compare)
            view.marshal(query=sq.parsed, results=sq.results, compare=sq.compare,
                parsed=sq.parsed, created=epoch(sq.created))

            viewarea.call("set_view", view)
            viewarea.context.update(view=view)

        filters = read_filters(query)
        compare_filters = read_filters(query, 'compare')

        qs = QuerySidebar(info=table_info, view=view, filters=filters or [],
            compare_filters=compare_filters, metadata=table_info)

        if compare_filters:
            qs.call('show_compare_filters')

        qs.call("supports_compare_filters", view.SUPPORT_COMPARE_QUERIES)
        if has_saved:
            qs.call("show_results")

        qs.set_ref("sidebar")
        qs.marshal(table=table, viewarea=viewarea, metadata=table_info)

        self.set_ref("querypage")


        user_button = UserButton()
        user_modal = UserModal(show_queries=True)

        self.marshal(sidebar=qs, table=table, user_modal=user_modal)

        self.context.update(
            table=table,
            sidebar=qs,
            user_button=user_button,
            user_modal=user_modal,
            viewarea=viewarea,
            table_selector=table_selector)

# TODO: place these on QueryPage?
@QuerySidebar.api
@rpc_login_required
def run_query(cls, table=None, query=None, viewarea=None, filters=[]):
    if not rbac.check("query", table):
        return {
            "error" : "You don't have access to this table"
        }

    user = flask_security.core.current_user.email
    # this is a name/value encoded array, unfortunately
    query = QuerySpec(query)
    query.add('table', table)
    query.add('filters', string_dict(filters))
    d = query.__makedict__()

    bs = backend.SybilBackend()

    ti = bs.get_table_info(table)

    view = query.get('view')
    VwClass = get_view_by_name(view)

    # this lets a class modify the parameters to a query using
    # its own custom params
    VwClass.add_custom_params(query)
    query.set('viewbase', VwClass.BASE)

    res = bs.run_query(table, query, ti)
    bs.log_query(user, query)

    cmp = None
    against = query.get('against', '')
    filters = query.get('filters')

    compare_filters = filters['compare']
    if against or len(compare_filters):
        compare_spec = QuerySpec(query.md)

        if against:
            now = time_to_seconds('now')
            compare_delta = time_to_seconds(against) - now

            # compare_delta is in ms
            query.set('compare_delta', compare_delta*1000)

            startms = query.get('start_ms')
            endms = query.get('end_ms')

            startms += compare_delta * 1000
            endms += compare_delta * 1000

            compare_spec.set('start_ms', startms)
            compare_spec.set('end_ms', endms)


        compare_spec.set('filters', { "query" : compare_filters })

        cmp = bs.run_query(table, compare_spec, ti)
        query.set('compare_mode', 1)

        bs.log_query(user, compare_spec)

    sq = results.save_for_user(flask_security.core.current_user, query, res, cmp)

    d["h"] = sq.hashid

    v = VwClass()
    v.context.update(query=sq.parsed, results=sq.results, metadata=ti, compare=sq.compare)
    v.marshal(query=sq.parsed, results=sq.results, compare=sq.compare, metadata=ti, parsed=epoch(sq.created))

    if viewarea:
        if not res:
            viewarea.call("no_results", {
                "errmsg" : "No results found for this query"
            });
        viewarea.html(v.render())
        # viewarea.call("set_view", v)


    d['filters'] = json.dumps(d['filters'])
    queryUrl = unquote_plus(flask.url_for('get_view', **d))
    return {
        "queryUrl": queryUrl,
        "res" : res,
        "cmp" : cmp,
        "query" : d
    }

@QuerySidebar.api
@rpc_login_required
def update_controls(cls, table=None, view=None, query=None, viewarea=None, filters=[]):
    pr = presenter.GetPresenter(table)

    bs = backend.SybilBackend()
    ti = bs.get_table_info(table)

    query = QuerySpec(query)

    VwClass = get_view_by_name(view)

    v = VwClass()
    v.context.update(metadata=ti, presenter=pr, query=query)
    query_filters= filters['query']

    qs = QuerySidebar(view=v, presenter=pr, query=query, filters=query_filters, metadata=ti)
    qs.__prepare__()
    qs.nomarshal()

    has_compare_filters = 'compare' in filters and len(filters['compare']) > 0
    if v.SUPPORT_COMPARE_QUERIES and has_compare_filters:
        cls.call("show_compare_filters")
    else:
        cls.call("hide_compare_filters")
    cls.call("supports_compare_filters", v.SUPPORT_COMPARE_QUERIES)

    # we undelegate our events because we are about to replace ourself
    # with the same component
    cls.html(qs.context.querycontrols.render(), selector=".querycontrols")

@QueryPage.api
@rpc_login_required
def get_saved_queries(cls, table=None):
    user = flask_security.core.current_user

    if table:
        recent_queries = results.get_for_user(user, table)

    return {
        "recent" : recent_queries,
        "table" : table

    }



from __future__ import absolute_import
import pudgy
from .components import *
from .query_sidebar import QuerySidebar
from .user_button import UserButton, UserModal

# old snorkel components are prefixed with sf
from . import sf_components as sf

class TourCarousel(pudgy.BackboneComponent, pudgy.JinjaComponent):
    def __prepare__(self):
        from . import tour
        self.context.slides = tour.slides

def install(app):
    pudgy.use_jquery()
    pudgy.add_to_prelude("bootstrap", os.path.join(app.static_folder, "vendor/bootstrap.min.js"))
    pudgy.add_prelude_line("require('bootstrap')");
    app.after_request(pudgy.compress_request)



# This is an example of how we would configure snorkel
# from the outside
# if we are outside snorkel package, it would be:

# from snorkel import web
# from snorkel.presenter import DatasetPresenter, RegisterPresenter


import os
from . import web
from .presenter import DatasetPresenter, RegisterPresenter, ConfigureTable

from .views import ViewSeparator

from .plugins.snorkel_basic_views import TableView, TimeView, DistView, SamplesView
from .plugins.snorkel_basic_views import AreaView, ScatterView, BarView, GroupedDist
from .plugins.snorkel_advanced_views import TimelineView, OverviewView, ForecastView, DrilldownView
from .plugins.snorkel_advanced_views import DigraphView, WecoView

def configure_presenters():
    default_presenter = DatasetPresenter()
    default_presenter.set_views([
        TableView,
        TimeView,
        DistView,
        SamplesView,
        TimelineView,
        OverviewView,
        ViewSeparator,
        AreaView,
        BarView,
        ScatterView,
        GroupedDist,
        ViewSeparator,
        DigraphView,
        ViewSeparator,
        WecoView,
        ForecastView,
        DrilldownView

    ])
    RegisterPresenter(".*", default_presenter)


def main():
    configure_presenters()

    PROFILE="PROFILE" in os.environ
    if PROFILE:
        from werkzeug.contrib.profiler import ProfilerMiddleware
        web.app.config['PROFILE'] = True
        web.app.wsgi_app = ProfilerMiddleware(web.app.wsgi_app, restrictions = [30])

    web.app.run(port=os.environ.get("PORT", 2333), use_reloader=False, debug=PROFILE)

if __name__ == "__main__":
    main()



import pudgy

from snorkel.views import ViewBase
from snorkel.components import *
import os

import numbers

TIME_SLICE_OPTIONS = [
    "auto",
    ("1 min", 60),
    ("5 min", 60 * 5),
    ("10 min", 60 * 10),
    ("30 min", 60 * 30),
    ("1 hour", 60 * 60),
    ("3 hours", 60 * 60 * 3),
    ("6 hours", 60 * 60 * 6),
    ("12 hours", 60 * 60 * 12),
    ("daily", 60 * 60 * 24),
    ("7 days", 60 * 60 * 24 * 7),
    ("30 days", 60 * 60 * 24 * 30),
    ("yearly", 60 * 60 * 24 * 365),
]


class nvd3(OldSnorkelComponent):
    pass

class TimeView(ViewBase, pudgy.JSComponent):
    NAME="time"
    BASE="time"
    DISPLAY_NAME="Time View"
    BASE_DIR=os.path.dirname(__file__)
    SUPPORT_COMPARE_QUERIES=True



    def add_time_series_controls(self, controls):
        time_slice = Selector(
            name="time_bucket",
            options=TIME_SLICE_OPTIONS,
            selected=self.context.query.get("time_bucket"))

        controls.append(ControlRow("time_bucket", "Time Slice", time_slice))

        normalize = Selector(
            name="time_divisor",
            options=[ "", "hour", "minute" ],
            selected=self.context.query.get("time_divisor"))
        controls.append(ControlRow("time_divisor", "Normalize", normalize))



    def get_controls(self):
        controls = []

        self.add_time_controls(controls)
#        self.add_time_comparison(controls)

        self.add_time_series_controls(controls)

        self.add_groupby_selector(controls)
        self.add_limit_selector(controls)

        self.add_metric_selector(controls)
        self.add_fields_selector(controls)
        self.add_go_button(controls)

        return controls

    def __prepare__(self):
        pass



from .models import SavedQuery
from pudgy.util  import getrandhash
from . import fastjson as json
from . import presenter

def save_for_user(user, params, results, compare={}):
    table = params.get('table')

    table = presenter.GetRealTable(table)

    params = params.__json__()
    hashid = getrandhash(json.dumps(params))
    hashid = hashid[:10]
    sq = SavedQuery.create(results=results, user=user.id, parsed=params, hashid=hashid,
        table=table, compare=compare or {})

    return sq

def get_for_user(user, table, limit=30):
    table = presenter.GetRealTable(table)
    res = SavedQuery \
        .select(
            SavedQuery.created,
            SavedQuery.parsed,
            SavedQuery.hashid,
            SavedQuery.user
        ).limit(limit) \
        .order_by(-SavedQuery.created) \
        .where(SavedQuery.user == user.id, SavedQuery.table == table)

    res = res.dicts()

    return list(reversed(res))

def get_by_hashid(hashid):
    return list(SavedQuery.select().where(SavedQuery.hashid == hashid).dicts())



import pudgy
import os

from snorkel.components import Selector, ControlRow
from snorkel.views import ViewBase

def make_dict(arr):
    return dict([(w,w) for w in arr])

class OverviewView(ViewBase, pudgy.JSComponent, pudgy.SassComponent):
    NAME="overview"
    BASE="samples"
    DISPLAY_NAME="Overview View"
    BASE_DIR=os.path.dirname(__file__)

    def get_controls(self):
        controls = []

        self.add_time_controls(controls)

        self.add_limit_selector(controls)

        self.add_go_button(controls)

        return controls




import flask_security
from .components import UIComponent
import pudgy

class UserButton(UIComponent, pudgy.MustacheComponent):
    def __prepare__(self):
        self.context.user = flask_security.core.current_user

class UserModal(UIComponent, pudgy.JinjaComponent,
        pudgy.ServerBridge, pudgy.SassComponent):
    def __prepare__(self):
        self.context.user = flask_security.core.current_user




import pudgy
import os

from snorkel.components import Selector, ControlRow
from snorkel.views import ViewBase

import sys

from ..snorkel_basic_views import TimeView

def make_dict(arr):
    return dict([(w,w) for w in arr])

# this line lets us access TimeView's modules via prefix
pudgy.components.add_dirhash_alias("timeview", TimeView)
class WecoView(TimeView, pudgy.JSComponent, pudgy.SassComponent):
    NAME="weco"
    BASE="time"
    DISPLAY_NAME="Weco View"
    BASE_DIR=os.path.dirname(__file__)




from flask import Flask, render_template, redirect, url_for
from werkzeug import Response
from flask_peewee.db import Database
from peewee import *
from flask_security import Security, PeeweeUserDatastore
from flask_security import login_required, login_user, logout_user

from flask_security.core import current_user, _security
from flask_dance.contrib.google import make_google_blueprint, google


from .models import User, Role, UserRoles, UserToken, userdb
from . import oauth, config, rbac

import random, sys, os

# Uses our own authentication tokens instead of the signed ones from
# flask_security because the verification of signatures takes too long
def _request_loader(request):
    header_key = _security.token_authentication_header
    args_key = _security.token_authentication_key
    header_token = request.headers.get(header_key, None)
    token = request.args.get(args_key, header_token)
    if request.is_json:
        data = request.get_json(silent=True) or {}
        if isinstance(data, dict):
            token = data.get(args_key, token)

    try:
        token = UserToken.select().where(UserToken.token == token).first()
        user = token.user
        login_user(user)
        return user
    except Exception as e:
        pass
    return _security.login_manager.anonymous_user()

import flask
def install(app):
    oauth.install(app)

    app.config['SECRET_KEY'] = config.SECRET_KEY
    app.config['SECURITY_PASSWORD_SALT'] = config.SECURITY_PASSWORD_SALT
    app.config['SECURITY_CHANGEABLE'] = True

    db = userdb

    # Setup Flask-Security
    user_datastore = PeeweeUserDatastore(db, User, Role, UserRoles)
    security = Security(app, user_datastore)
    app.login_manager.request_loader(_request_loader)


    # Views
    @app.route('/')
    @needs_login
    def home():
        return render_template('index.html')


def check_auth():
    auth = False
    if current_user.is_authenticated:
        auth = True
    else:
        if google.authorized:
            err = False
            try:
                resp_json = google.get("/oauth2/v2/userinfo").json()
                if "error" in resp_json:
                    err = True
            except Exception as e:
                err = True

            if err:
                oauth.logout_oauth()

    return auth

def needs_login(func):
    def wrapped_func(*args, **kwargs):

        try:
            auth = check_auth()
        except Exception as e:
            auth = False

        if not auth:
            return redirect(url_for("security.login"))

        return func(*args, **kwargs)

    wrapped_func.__name__ = func.__name__
    return wrapped_func

# if our login_required returns a Response with code 300, we know to reload the
# page
def rpc_login_required(func):
    f = needs_login(func)
    def wrapped_func(*args, **kwargs):

        r = f(*args, **kwargs)
        if isinstance(r, Response):
            # if we have a redirect, we need to translate it to the grpc framework
            if r.status_code in [301, 302, 303, 305, 307]:
                flask.request.pudgy.activations.append("window.location.reload()")

                # get and unset all the flashed messages
                # TODO: this is a temporary workaround for duplicate messages
                flashed = set(flask.get_flashed_messages(with_categories=True))
                return { "_status" : r.status, "_nextUrl" : r.location }

        return r

    wrapped_func.__name__ = func.__name__

    return wrapped_func

if __name__ == '__main__':
    app = Flask(__name__)
    install(app)
    app.run()



import pudgy

from .time import TimeView
from .dist import DistView
from .table import TableView
from .samples import SamplesView
from .area import AreaView
from .scatter import ScatterView
from .bar import BarView
from .grouped_dist import GroupedDist

pudgy.add_dirhash_alias("snorkel", TimeView)



from __future__ import absolute_import
import pudgy

from snorkel.views import ViewBase
from snorkel.components import *

from .dist import DistView
import os

class GroupedDist(DistView):
    NAME="grouped_dist"
    BASE="dist"
    DISPLAY_NAME="Grouped Dist"
    BASE_DIR=os.path.dirname(__file__)
    SUPPORT_COMPARE_QUERIES = False

    def get_controls(self):
        controls = super(GroupedDist, self).get_controls()

        self.add_groupby_selector(controls)

        return controls



from __future__ import absolute_import
import pudgy

from snorkel.views import ViewBase
from snorkel.components import *

from .table import TableView
import os

class BarView(TableView):
    NAME="bar"
    BASE="table"
    DISPLAY_NAME="Bar View"
    BASE_DIR=os.path.dirname(__file__)



import sys

from .. import components
from .view import ViewBase, ViewSeparator
from .view import get_view_by_name, get_column_types

## SETUP PLUGINS HERE
import addict

if not "snorkel" in sys.modules:
    snorkel = addict.Dict({ "views" : sys.modules[__name__], "components" : components})
    sys.modules["snorkel"] = snorkel
    sys.modules["snorkel.views"] = snorkel.views
    sys.modules["snorkel.components"] = snorkel.components



import pudgy

from snorkel.views import ViewBase
from snorkel.components import *

from ..snorkel_basic_views import TableView
import os

class DigraphView(TableView):
    NAME="digraph"
    BASE="table"
    DISPLAY_NAME="DiGraph View"
    BASE_DIR=os.path.dirname(__file__)
    SUPPORT_COMPARE_QUERIES=False

    @classmethod
    def add_custom_params(self, query_spec):
        dim_one = query_spec.get("dim_one")
        dim_two = query_spec.get("dim_two")
        groupby = query_spec.get_groupby()
        groupby.extend([dim_one, dim_two])
        query_spec.setlist('groupby[]', groupby)

    def add_graph_controls(self, controls):
        def make_dict(arr):
            return dict([(w,w) for w in arr])

        groups = make_dict(self.context.metadata["columns"]["strs"])
        prev_node = Selector(
            name="dim_one",
            options=groups,
            selected=self.context.query.get('dim_one', ''))

        next_node = Selector(
            name="dim_two",
            options=groups,
            selected=self.context.query.get('dim_two', ''))

        controls.append(ControlRow("prev_node", "Prev Node", prev_node))
        controls.append(ControlRow("next_node", "Next Node", next_node))


    def get_controls(self):
        controls = []
        self.add_time_controls(controls)

        self.add_groupby_selector(controls)
        self.add_limit_selector(controls)

        self.add_graph_controls(controls)

        self.add_go_button(controls)

        return controls



from __future__ import print_function
from peewee import *
from playhouse.sqlite_ext import *

from flask_security import RoleMixin, UserMixin

import os

from playhouse.sqliteq import SqliteQueueDatabase
from playhouse.migrate import SqliteMigrator, migrate

DB_DIR="sdb"
userdb = SqliteDatabase(os.path.join(DB_DIR, 'users.db'))
querydb = SqliteDatabase(os.path.join(DB_DIR, 'queries.db'))

class QueryModel(Model):
    class Meta:
        database = querydb

class UserModel(Model):
    class Meta:
        database = userdb

class SavedQuery(QueryModel):
    user = IntegerField(index=True)
    table = CharField(index=True)

    hashid = CharField(index=True)
    created = TimestampField(index=True, utc=True)
    updated = TimestampField(utc=True)

    results = JSONField()
    compare = JSONField()
    parsed = JSONField()

    def toObject(self):
        return {
            "results" : self.results,
            "parsed" : self.parsed,
            "created" : self.created

        }

    class Meta:
        database = querydb

class Role(UserModel, RoleMixin):
    name = CharField(unique=True)
    description = TextField(null=True)

class User(UserModel, UserMixin):
    email = TextField()
    password = TextField()
    active = BooleanField(default=True)
    confirmed_at = DateTimeField(null=True)

    def get_auth_token(self):
        tokens = list(self.get_tokens())
        if not tokens:
            UserToken.create_for_user(self)

        return self.tokens[0].token

    def get_tokens(self):
        return self.tokens

class UserToken(UserModel):
    user = ForeignKeyField(User, related_name='tokens')
    token = TextField(unique=True)
    active = BooleanField(default=True)

    @classmethod
    def create_for_user(cls, user):
        import uuid
        randid = uuid.uuid4()
        token = UserToken.create(token=randid, user=user)
        return token



class UserRoles(UserModel):
    # Because peewee does not come with built-in many-to-many
    # relationships, we need this intermediary class to link
    # user to roles.
    user = ForeignKeyField(User, related_name='roles')
    role = ForeignKeyField(Role, related_name='users')
    name = property(lambda self: self.role.name)
    description = property(lambda self: self.role.description)

def create_db_if_not():
    try:
        os.makedirs(DB_DIR)
    except Exception as e:
        pass

    User._meta.database.connect()
    for c in [SavedQuery, User, UserToken, UserRoles]:
        c._meta.database.create_tables([c])

    # query DB migrations
    # not necessarily needed
    migrator = SqliteMigrator(querydb)
    try:
        migrate(
            migrator.add_column('savedquery', 'compare', JSONField(default='')),
        )
    except Exception as e:
        print(e)

if __name__ == "__main__":
    if "RESET" in os.environ:
        create_db_if_not()



from __future__ import print_function
import os
from subprocess import Popen, PIPE, check_output
from flask_security import current_user

import subprocess
import shlex
import pudgy
import time

import sys

from .. import fastjson as json

from .. import presenter

USING_MSYBIL = False

# TODO: enable REMOTE_INGEST via config variable
ENABLE_REMOTE_INGEST=False
MSYBIL_BIN = os.path.join(os.path.dirname(__file__), "msybil.py")
MSYBIL_INGEST_BIN = os.path.join(os.path.dirname(__file__), "msybil_ingest.py")
SYBIL_BIN = os.path.join(os.path.dirname(__file__), "bin", "sybil")
DEBUG="DEBUG" in os.environ

if "MSYBIL" in os.environ:
    # TODO: make msybil.py read off MSYBIL environment variable instead of stdin
    print(" s Using Multisybil", file=sys.stderr)

    SYBIL_BIN = MSYBIL_BIN
    USING_MSYBIL = True



from collections import defaultdict
from .backend import Backend

from ..util import time_to_seconds, time_delta_to_seconds

import re

def enquote(s):
    return '%s' % s

def get_time_col(md):
    for c in ["time", "integer_time", "timestamp"]:
        if c in md["col_types"]:
            return c

    return "time"

# group 1 is the agg (avg, sum, etc)
# group 2 is the col name
col_re = re.compile("(.*)\((.*)\)")
def extract_field(col):
    m = col_re.match(col)
    if m:
        return m.group(2)

    return col

def run_query_command(cmd_args):
    init_cmd_args = [SYBIL_BIN, "query", "-json"]
    init_cmd_args.extend(["--read-log"])
    init_cmd_args.extend(["--cache-queries"])
    init_cmd_args.extend(["--field-separator=%s" % FIELD_SEPARATOR])
    init_cmd_args.extend(["--filter-separator=%s" % FILTER_SEPARATOR])
    init_cmd_args.extend(cmd_args)

    init_cmd_args = [str(a).encode('ascii', errors='replace') for a in init_cmd_args]

    ret = run_command(init_cmd_args)
    return json.loads(ret)

def run_command(cmd_args, stdin=b""):
    cmd_args = list(map(lambda w: w.decode("utf-8"), cmd_args))
    print("RUNNING COMMAND", " ".join(cmd_args))

    if isinstance(stdin, str):
        stdin = stdin.encode("utf-8")

    p = Popen(cmd_args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    stdout, stderr = p.communicate(stdin)
    if DEBUG:
        print(stderr)

    return stdout.decode("utf-8")

FIELD_SEPARATOR=chr(30)
FILTER_SEPARATOR=chr(31)
#FIELD_SEPARATOR=","
#FILTER_SEPARATOR=":"

def estimate_time_buckets(query_interval, buckets):
    best_bucket_count = buckets or 1000;
    min_intervals = [ 5, 10, 30, 60, 120, 360, 720, 1440 ];
    interval_ms = 0

    i = 0
    while i < len(min_intervals):
        interval = min_intervals[i] * 60 * 1000;
        interval_ms = min_intervals[i] * 60;
        if (query_interval / interval < best_bucket_count):
            break;

        i += 1

    return interval_ms;


class SybilQuery(object):
    def run_query(self, table, query_spec, metadata):
        self.metadata = metadata
        table = presenter.GetRealTable(table)

        self.table = table
        self.query = query_spec

        view = query_spec.get('viewbase') or query_spec.get('view')

        if view == 'table':
            return self.run_table_query(table, query_spec)

        elif view == 'time':
            return self.run_time_query(table, query_spec)
        elif view == 'dist':
            return self.run_dist_query(table, query_spec)
        elif view == 'samples':
            return self.run_samples_query(table, query_spec)
        else:
            raise Exception("unrecognized view", view)

    def add_op(self, query_spec, cmd_args):
        # if only one metric is specified:
        op = query_spec.get_metric() or "count"
        if op == "Distinct":
            cmd_args.extend(["-op", "distinct"])
        elif op[0].lower() == "p": # p25, p50, etc
            cmd_args.extend(["-op", "hist"])
        else: # we are doing an avg, sum or count query
            pass


    def add_weight_col(self, query_spec, cmd_args):
        weight_col = None
        for c in [ "weight" ]:
            if c in self.metadata["col_types"]:
                weight_col = c

        if weight_col:
            cmd_args.extend(["-weight-col", weight_col])
            query_spec.set('weight_col', weight_col)


    def add_group_by(self, query_spec, cmd_args):
        groupby = query_spec.get_groupby()
        if groupby:
            cmd_args.append("-group")
            cmd_args.append(FIELD_SEPARATOR.join(groupby))

    def add_fields(self, query_spec, cmd_args):
        all_fields = []

        fields = query_spec.get_fields()
        if fields:
            all_fields.extend(fields)

        custom_fields = query_spec.getlist("custom_fields[]")
        if custom_fields:
            all_fields.extend([extract_field(c) for c in custom_fields])
            cmd_args.extend(['-op', 'hist'])
            cmd_args.extend(["-loghist"])

        field = query_spec.get("field")
        if field:
            all_fields.append(field)


        if all_fields:
            cmd_args.append("-int")
            cmd_args.append(FIELD_SEPARATOR.join(all_fields))

        query_spec.set('all_fields', all_fields)

    def add_filters(self, query_spec, cmd_args):
        from ..views import get_column_types


        # add time filters
        start_ms = query_spec.get('start_ms', '')
        end_ms = query_spec.get('end_ms', '')

        if start_ms:
            start_s = int(start_ms) / 1000
        else:
            start = query_spec.get('start', "-1 week")
            start_s = time_to_seconds(start)

        if end_ms:
            end_s = int(end_ms) / 1000
        else:
            end = query_spec.get('end', "now")
            end_s = time_to_seconds(end)


        custom_end = query_spec.get('custom_end', '')
        custom_start = query_spec.get('custom_start', '')

        if custom_start:
            start_s = time_to_seconds(custom_start)

        if custom_end:
            end_s = time_to_seconds(custom_end)

        query_spec.set('start_ms', start_s * 1000)
        query_spec.set('end_ms', end_s * 1000)


        query_interval = abs(end_s - start_s) * 1000;
        time_bucket = query_spec.get("time_bucket", "")

        if not time_bucket or time_bucket == "auto":
            time_bucket = estimate_time_buckets(query_interval, 800)

        query_spec.set('time_bucket', time_bucket)



        md = self.metadata
        fields, types = get_column_types(md)

        time_col = get_time_col(md)

        filters = query_spec.get('filters')['query']

        int_filters = defaultdict(list)
        str_filters = defaultdict(list)

        for f in filters:
            col, op, val = f
            if val == "":
                continue

            if op[0] != "$":
                continue

            op = op[1:]

            if op == "regex":
                # this is for compatibility with snorkelv1 filters
                op = "re"

            if col in types:
                tp = types[col]



                if tp == "string":
                    str_filters[col].append((op, val))
                elif tp == "integer":
                    int_filters[col].append((op, val))



        int_filters[time_col].append(("gt", start_s))
        int_filters[time_col].append(("lt", end_s))

        filter_strs = []
        for col in int_filters:
            for f in int_filters[col]:
                filter_strs.append(FILTER_SEPARATOR.join(map(str, [col, f[0], f[1]])))

        if filter_strs:
            cmd_args.extend(["-int-filter", enquote(FIELD_SEPARATOR.join(filter_strs))])

        filter_strs = []
        for col in str_filters:
            for f in str_filters[col]:
                filter_strs.append(FILTER_SEPARATOR.join(map(str, (col, f[0], f[1]))))

        if filter_strs:
            cmd_args.extend(["-str-filter", enquote(FIELD_SEPARATOR.join(map(str, filter_strs)))])

    def add_limit(self, query_spec, cmd_args):
        # TODO: limit should depend on the view. time series limit defaults to 10 or 20
        limit = query_spec.get("limit", 100)
        cmd_args.extend(["-limit", limit])


    def run_table_query(self, table, query_spec):
        cmd_args = [ "-table", table ]

        self.add_op(query_spec, cmd_args)
        self.add_weight_col(query_spec, cmd_args)
        self.add_group_by(query_spec, cmd_args)
        self.add_fields(query_spec, cmd_args)
        self.add_filters(query_spec, cmd_args)
        self.add_limit(query_spec, cmd_args)

        return run_query_command(cmd_args)


    def run_dist_query(self, table, query_spec):
        cmd_args = [ "-table", table ]

        cmd_args.extend(['-op', 'hist', '-loghist'])

        self.add_weight_col(query_spec, cmd_args)
        self.add_group_by(query_spec, cmd_args)
        self.add_fields(query_spec, cmd_args)
        self.add_filters(query_spec, cmd_args)
#        self.add_hist_buckets(query_spec, cmd_args)

        return run_query_command(cmd_args)

    def run_time_query(self, table, query_spec):

        md = self.metadata
        time_col = get_time_col(md)
        cmd_args = [ "-table", table, "-time-col", time_col, "-time"]

        self.add_op(query_spec, cmd_args)
        self.add_weight_col(query_spec, cmd_args)
        self.add_group_by(query_spec, cmd_args)
        self.add_fields(query_spec, cmd_args)
        self.add_filters(query_spec, cmd_args)
        self.add_limit(query_spec, cmd_args)

        time_bucket = query_spec.get("time_bucket", 300)
        if time_bucket != "auto":
            cmd_args.extend(["-time-bucket", str(time_bucket) ])

        return run_query_command(cmd_args)

    def run_samples_query(self, table, query_spec):
        cmd_args = [ "-table", table, "-samples" ]
        self.add_filters(query_spec, cmd_args)
        self.add_limit(query_spec, cmd_args)
        return run_query_command(cmd_args)


@pudgy.util.memoize
def get_table_info(table, ts=None):
    from ..views import get_column_types
    table = presenter.GetRealTable(table)
    meta = run_query_command(["-table",  table, "-info"])
    fields, types = get_column_types(meta)
    meta["col_types"] = types
    return meta

class SybilBackend(Backend):
    def clear_cache(self, table=None):
        if table:
            if table in get_table_info.cache:
                del get_table_info.cache[table]
        else:
            get_table_info.cache.clear()

    def list_tables(self):
        tables = run_query_command(["-tables"])
        tables = map(presenter.GetTableName, tables)
        tables = list(filter(presenter.IsTableVisible, tables))
        tables.sort()
        return tables


    def get_table_info(self, table):
        return get_table_info(table)

    def log_query(self, user, query_spec):
        sample_data = {
          "start_time": query_spec.get('start_ms'),
          "end_time": query_spec.get('end_ms'),

          "weight_col": query_spec.get('weight_col') or "unspecified",
          "view": query_spec.get('view'),
          "start": query_spec.get('start') or query_spec.get('custom_start'),
          "end": query_spec.get('end') or query_spec.get('custom_end'),
          "table": query_spec.get('table'),
          "user": user,
          "time" : time.time(),
          "cols": query_spec.get_fields(),
          "dims": query_spec.get_groupby() }

        if not USING_MSYBIL or ENABLE_REMOTE_INGEST:
            self.ingest("slite@queries", [ sample_data ], log_ingest=False)

    def run_query(self, table, query_spec, metadata):
        q = SybilQuery()
        return q.run_query(table, query_spec, metadata)

    # TODO: for msybil, randomly pick a server and send samples to it
    def ingest(self, table, samples, log_ingest=True):
        if USING_MSYBIL:
            if ENABLE_REMOTE_INGEST:
                # TODO: save the samples locally. if the ingestion succeeds
                # remove the file. else we need to re-enqueue for ingestion
                cmd_args = [MSYBIL_INGEST_BIN, table]
            else:
                raise Exception("REMOTE INGESTION NOT ENABLED FOR MSYBIL")
        else:
            cmd_args = [SYBIL_BIN, "ingest", "-table", table]

        print("INGESTING %s SAMPLES INTO %s" % (len(samples), table))
        cmd_args = map(lambda w: w.encode("utf-8"), cmd_args)
        run_command(cmd_args, stdin=json.dumps(samples))

        if log_ingest:
            user = "anonymous"
            try:
                user = current_user.email
            except:
                pass
            self.ingest("slite@ingest", [{
                "table": table,
                "count" : len(samples),
                "time" : int(time.time()),
                "user" : user
            }], log_ingest=False)

if not os.path.exists(SYBIL_BIN):
    SYBIL_BIN = run_command(["which", "sybil"]).strip()

if __name__ == "__main__":
    print(time_to_seconds('-1 week'))
    print(time_delta_to_seconds('-1 week'))



import pudgy
import os

from snorkel.components import Selector, ControlRow
from snorkel.views import ViewBase

import sys

from ..snorkel_basic_views import TimeView

def make_dict(arr):
    return dict([(w,w) for w in arr])

# this line lets us access TimeView's modules via prefix
pudgy.components.add_dirhash_alias("timeview", TimeView)
class ForecastView(TimeView, pudgy.JSComponent, pudgy.SassComponent):
    NAME="forecast"
    BASE="time"
    DISPLAY_NAME="Forecasting"
    BASE_DIR=os.path.dirname(__file__)

    def add_forecast_controls(self, controls):
        groups = make_dict(self.context.metadata["columns"]["strs"])
        model_type = Selector(
            name="model_type",
            options={ "additive" : "additive", "multiplicative" : "multiplicative"},
            selected=self.context.query.get("model_type"))

        controls.append(ControlRow("model_type", "Model Type", model_type))

        iter_mode = Selector(
            name="iter_mode",
            options={"fast" : "fast", "slow" : "thorough"},
            selected=self.context.query.get("iter_mode"))

        controls.append(ControlRow("iter_mode", "Exploration", iter_mode))

        componentize = Selector(
            name="componentize",
            options={"" : "hide", "componentize" : "show"},
            selected=self.context.query.get("componentize"))

        controls.append(ControlRow("componentize", "Components", componentize))

    def get_controls(self):
        controls = []

        self.add_time_controls(controls)

        self.add_time_series_controls(controls)

        self.add_groupby_selector(controls)
        self.add_limit_selector(controls)

        self.add_forecast_controls(controls)

        self.add_metric_selector(controls)
        self.add_fields_selector(controls)
        self.add_go_button(controls)

        return controls



import subprocess
from . import fastjson as json


# time translation command is:
# `date -d "<str>" +%s`
def time_to_seconds(timestr):
    cmd_args = ["date", "-d", timestr, "+%s"]
    try:
        output = subprocess.check_output(cmd_args)
    except:
        raise Exception("Unknown time string: ", timestr)
    return int(output)

def time_delta_to_seconds(timedelta):
    now = time_to_seconds("now")
    then = time_to_seconds(timedelta)

    return now - then

def return_json(d):
    return json.dumps(d), 200, {"ContentType" : "application/json"}

# from https://stackoverflow.com/questions/1254454/fastest-way-to-convert-a-dicts-keys-values-from-unicode-to-str
import collections

try:
  basestring
except NameError:
  basestring = str

def convert(data):
    if isinstance(data, basestring):
        return str(data)
    elif isinstance(data, collections.Mapping):
        return dict(map(convert, data.items()))
    elif isinstance(data, collections.Iterable):
        return type(data)(map(convert, data))
    else:
        return data

string_dict = convert



import pudgy
import os

from snorkel.components import Selector, ControlRow
from snorkel.views import ViewBase

def make_dict(arr):
    return dict([(w,w) for w in arr])

class ScatterView(ViewBase, pudgy.JSComponent):
    NAME="scatter"
    BASE="samples"
    DISPLAY_NAME="Scatter View"
    BASE_DIR=os.path.dirname(__file__)

    def add_scatter_controls(self, controls):
        fields = make_dict(self.context.metadata["columns"]["ints"])
        field = Selector(
            name="field",
            options=fields,
            selected=self.context.query.get('field'))

        field2 = Selector(
            name="field_two",
            options=fields,
            selected=self.context.query.get('field_two'))

        reverse_axis = Selector(
            name="reverse_axis",
            options={
                "Reverse Both " : "reverse_b",
                "Reverse Y Axis" : "reverse_y",
                "Reverse X Axis" : "reverse_x",
                "Normal" : "none",
            },
            selected=self.context.query.get('reverse_axis', "none"))


        controls.append(ControlRow("field", "Field", field))
        controls.append(ControlRow("field_two", "Field (2)", field2))
        controls.append(ControlRow("reverse_axis", "Reverse Axis", reverse_axis))



    def get_controls(self):
        controls = []

        self.add_time_controls(controls)
#        self.add_time_comparison(controls)

        self.add_scatter_controls(controls)

        self.add_groupby_selector(controls)
        self.add_limit_selector(controls)

        self.add_go_button(controls)

        return controls



from __future__ import print_function
import pudgy
from ..components import *
from .. import backend

from ..components import OldSnorkelComponent

import werkzeug
import os
import jinja2

def make_dict(arr):
    return dict([(w,w) for w in arr])

def fieldname(a, c):
    return "%s(%s)" % (a.strip(), c)

def displayname(a, c):
    return "%s(%s)" % (a.strip(), c)

START_TIME_OPTIONS = [
    "-1 hour",
    "-3 hours",
    "-6 hours",
    "-12 hours",
    "-1 day",
    "-3 days",
    "-1 week",
    "-2 weeks",
    "-3 weeks",
    "-1 month",
    "-2 months",
    "-3 months",
    "-6 months",
    "-1 year"
]

END_TIME_OPTIONS = [
    "Now",
    "-1 hour",
    "-3 hours",
    "-6 hours",
    "-12 hours",
    "-1 day",
    "-3 days",
    "-1 week",
    "-2 weeks",
    "-1 month",
]

AGAINST_TIME_OPTIONS = [
    "",
    "-1 hour",
    "-3 hours",
    "-6 hours",
    "-12 hours",
    "-1 day",
    "-3 days",
    "-1 week",
    "-2 weeks",
]

VIEW_OPTIONS = []


from ..components import sf
FilterRow = sf.filter_row

def get_column_types(md):
    types = {}
    fields = {}

    for field in md['columns']['strs']:
        fields[field] = field
        types[field] = 'string'

    for field in md['columns']['ints']:
        fields[field] = field
        types[field] = 'integer'

    return fields, types

@pudgy.Virtual
class ViewBase(pudgy.BackboneComponent):
    NAMESPACE = "views"
    DISPLAY_NAME=""
    NAME="ViewBase"

    SUPPORT_COMPARE_QUERIES=False

    @classmethod
    def get_display_name(self):
        return self.DISPLAY_NAME or self.NAME

    @classmethod
    def get_name(self):
        return "%s" % (self.NAME)

    @classmethod
    def add_custom_params(self, query_spec):
        return


    def add_time_controls(self, controls):

        has_custom = False
        if self.context.query.get('custom_start'):
            has_custom = True

        start_time = Selector(
            name="start",
            options=START_TIME_OPTIONS,
            selected=self.context.query.get('start', '-1 week'))

        end_time = Selector(
            name="end",
            options=END_TIME_OPTIONS,
            selected=self.context.query.get('end'))

        controls.append(ControlRow("start", "Start", start_time, hidden=has_custom))
        controls.append(ControlRow("end", "End", end_time, hidden=has_custom))

        custom_start = TextInput(
            name="custom_start",
            value=self.context.query.get('custom_start', '')
        )
        controls.append(ControlRow("custom_start", "Start", custom_start,
            hidden=not has_custom))

        custom_end = TextInput(
            name="custom_end",
            value=self.context.query.get('custom_end', '')
        )
        controls.append(ControlRow("custom_end", "End", custom_end,
            hidden=not has_custom))

        title = "use custom time"
        if has_custom:
            title = "quick select time"

        controls.append(jinja2.Markup("<div class='clearfix centered'><a href='#' class='mrl mtl custom_time_inputs' >%s</a></div>" % title))


        if self.SUPPORT_COMPARE_QUERIES:
            self.add_time_comparison(controls)

    def add_time_comparison(self, controls):
        against_time = Selector(
            name="against",
            options=AGAINST_TIME_OPTIONS,
            selected=self.context.query.get('against'))

        controls.append(ControlRow("against", "Against", against_time))

    def add_view_selector(self, controls):
        view_selector = Selector(
            name="view",
            options=self.context.presenter.get_views(),
            selected=self.context.query.get('view'))

        controls.append(ControlRow("view", "View", view_selector))

    def add_limit_selector(self, controls):
        limit_selector = TextInput(
            name="limit",
            value=self.context.query.get('limit', 10) or 10)

        controls.append(ControlRow("limit", "Limit", limit_selector))

    def add_groupby_selector(self, controls):
        groups = make_dict(self.context.metadata["columns"]["strs"])
        groupby = MultiSelect(
            name="groupby[]",
            options=groups,
            selected=self.context.query.get_groupby())

        print("GROUPS", self.context.query.get_groupby())
        controls.append(ControlRow("groupby[]", "Group By", groupby))

    def add_metric_selector(self, controls):

        has_custom = False
        if self.context.query.get_custom_fields():
            has_custom=True

        metric_selector = Selector(
            name="metric",
            options=self.context.presenter.get_metrics(),
            selected=self.context.query.get('metric'))

        controls.append(ControlRow("metric", "Metric", metric_selector, hidden=has_custom))


    def add_field_selector(self, controls):
        fields = make_dict(self.context.metadata["columns"]["ints"])
        fields = Selector(
            name="field",
            options=fields,
            selected=self.context.query.get('field'))
        controls.append(ControlRow("field", "Field", fields))

    def add_fields_selector(self, controls):
        has_custom = False
        custom_str = "use custom fields"
        if self.context.query.get_custom_fields():
            custom_str = "quick select fields"
            has_custom = True

        fields = make_dict(self.context.metadata["columns"]["ints"])
        fields = MultiSelect(
            name="fields[]",
            options=fields,
            selected=self.context.query.get_fields())
        controls.append(ControlRow("fields[]", "Fields", fields, hidden=has_custom))


        # add custom fields

        metrics = self.context.presenter.get_metrics()
        int_fields = self.context.metadata["columns"]["ints"]
        custom_fields = {}
        for m in metrics:
            m = m.lower()
            for c in int_fields:
                custom_fields[fieldname(m, c)] = displayname(m, c)

        cf_selector = MultiSelect(
            name="custom_fields[]",
            options=custom_fields,
            selected=self.context.query.get_custom_fields())
        controls.append(ControlRow("custom_fields[]", "Custom Fields", cf_selector, hidden=not has_custom))

        custom_toggler = "<div class='' style='text-align: center'><a href='#' class='custom_field_inputs mrl'>%s</a></div>" % custom_str
        controls.append(jinja2.Markup(custom_toggler))


    def add_go_button(self, controls):
        button = Button(name='Go', className='go')
        controls.append(button)

    def get_filters(self, filters):
        controls = []
        md = self.context.metadata

        fields, types = get_column_types(md)
        if not filters:
            # at least one default filter
            filters = [[ "", "", ""]]

        for f in filters:
            filter = FilterRow(fields=fields)
            filter.context.update(
                selected=f[0],
                op=f[1],
                value=f[2]
            )
            filter.marshal(types=types)
            controls.append(filter)

        return controls


    def get_view_control(self):
        controls = []

        self.add_go_button(controls)
        self.add_view_selector(controls)

        return controls

    def get_controls(self):
        controls = []

        self.add_time_controls(controls)

        self.add_groupby_selector(controls)
        self.add_limit_selector(controls)

        self.add_metric_selector(controls)
        self.add_fields_selector(controls)

        return controls

    def get_stats(self):
        pass

class ViewSeparator(pudgy.Component):
    DISPLAY_NAME = "--"
    NAME = "Separator"

    @classmethod
    def get_display_name(self):
        return self.DISPLAY_NAME or self.NAME

    @classmethod
    def get_name(self):
        return "%s" % (self.NAME)


def get_view_by_name(name):
    for cls in pudgy.util.inheritors(ViewBase):
        if cls.NAME.lower() == name.lower():
            return cls



from __future__ import print_function
from flask import session, abort, redirect
from flask_dance.consumer import oauth_authorized
from flask_dance.contrib.google import make_google_blueprint, google
from flask_security import login_user, logout_user
from .models import User, userdb
from . import config
import random
import requests
import sys


from .config import USE_GOOGLE_AUTH, AUTHORIZED_DOMAIN
def setup_google_blueprint(app):
    global USE_GOOGLE_AUTH

    if not USE_GOOGLE_AUTH:
        return

    google_bp = make_google_blueprint(
        client_id=config.GOOGLE_CLIENT_ID,
        client_secret=config.GOOGLE_SECRET,
        scope=["email"]
    )
    app.register_blueprint(google_bp)


    @oauth_authorized.connect_via(google_bp)
    def logged_in(blueprint, token):
        resp_json = google.get("/oauth2/v2/userinfo").json()
        email = resp_json["email"]
        hd = email.split("@")[-1]
        print("HD IS", hd)
        if AUTHORIZED_DOMAIN and hd != AUTHORIZED_DOMAIN:
            requests.post(
                "https://accounts.google.com/o/oauth2/revoke",
                params={"token": token["access_token"]}
            )
            session.clear()
            abort(403)

        pw = "%x" % int(random.random() * sys.maxsize)
        user, found = User.get_or_create(email=email,defaults={"password" : pw})
        user.save()

        login_user(user)

def logout_oauth():
    logout_user()
    try:
        if google.authorized:
            try:
                token = google.token["access_token"]
                resp = google.post(
                    "https://accounts.google.com/o/oauth2/revoke",
                    params={"token": token},
                    headers={"Content-Type": "application/x-www-form-urlencoded"}
                )
                print(resp.ok, resp.text)
            except Exception as e:
                print(e)
    except:
        pass

def install(app):
    app.jinja_env.globals.update(google_auth=USE_GOOGLE_AUTH)
    setup_google_blueprint(app)

    @app.route("/logout")
    def logout():
        logout_oauth()
        return redirect('/')



from __future__ import print_function
from __future__ import absolute_import
# TODO: put flask CLI commands in here.
import importlib
import sys
import os

import click
from flask import Flask
from getpass import getpass

from .web import app
from .models import User, UserToken

from flask_security.utils import encrypt_password, verify_password


@app.cli.command()
@click.argument('name')
def add_user(name):
    dir = os.path.join(os.path.dirname(__file__), "..", "src")
    sys.path.append(dir)

    from .models import User

    print("Adding user '%s'" % name)

    try:
        user = User.get(User.email == name)
        print("User '%s' already exists" % name)
    except User.DoesNotExist:
        pw = getpass()

        user = User.create(email=name, password=encrypt_password(pw))
        user.save()
        print("Created user", name)


@app.cli.command()
@click.argument('name')
def add_superuser(name):
    from flask_security.datastore import PeeweeUserDatastore
    from .models import User, Role, UserRoles, userdb


    user_datastore = PeeweeUserDatastore(userdb, User, Role, UserRoles)
    try:
        user = User.get(User.email == name)
        print("User '%s' already exists" % name)

        while True:
            r = raw_input("Make user a superuser? [y/N]").lower()
            if r == "y":
                role = user_datastore.find_role('superuser')
                if not role:
                    user_datastore.create_role('superuser')
                user_datastore.add_role_to_user(user, role)
                user.save()
                print("Set user '%s' to superuser" % name)
                break
            elif r == "n":
                break


    except User.DoesNotExist:
        pw = getpass()
        user = User.create(username=name, email=name, password=pw, roles=['superuser'])
        role = user_datastore.find_role('superuser')
        if not role:
            user_datastore.create_role('superuser')
        user.save()
        print("Created user", name)


@app.cli.command()
@click.argument('name')
def get_user_token(name):
    try:
        user = User.get(User.email == name)
        print(user.get_auth_token())

    except User.DoesNotExist:
        print("No such user '%s'" % name)



# google auth settings
USE_GOOGLE_AUTH = False
GOOGLE_CLIENT_ID=None
GOOGLE_CLIENT_SECRET=None
AUTHORIZED_DOMAIN=None # anyone can login from any domain

# Role Based Authentication
# this lets you choose which users see which tables
# look at default.rbac for how to set this file up
AUTHORIZED_USERS=None

# flask security settings
SECRET_KEY='changme'
SECURITY_PASSWORD_SALT= b"changemeharder"

# show the default welcome page
SHOW_TOUR=False



from .components import OldSnorkelComponent

class filter_row(OldSnorkelComponent):
    pass

class tabs(OldSnorkelComponent):
    pass

class modal(OldSnorkelComponent):
    # for our modal, we want it to close, so
    # we don't force "display: inherit !important"
    @classmethod
    def add_display_rules(cls, data):
        return data

class timeago(OldSnorkelComponent):
    pass

class tablesorter(OldSnorkelComponent):
    pass

class table_popover(OldSnorkelComponent):
    pass

class query_details(OldSnorkelComponent):
    pass

class query_tile(OldSnorkelComponent):
    pass



import pudgy

import os

@pudgy.Virtual
class OldSnorkelComponent(pudgy.SuperfluousComponent):
    NAMESPACE = "sf"

class UIComponent(pudgy.Component):
    NAMESPACE = "ui"

class Button(UIComponent, pudgy.MustacheComponent):
    pass

class TextInput(UIComponent, pudgy.MustacheComponent):
    pass

class Table(UIComponent, pudgy.MustacheComponent, pudgy.SassComponent):
    pass

class Selector(UIComponent, pudgy.MustacheComponent, pudgy.SassComponent):
    def __init__(self, *args, **kwargs):
        super(Selector, self).__init__(*args, **kwargs)

        self.context.name = kwargs.get('name')
        self.context.options = []

        options = kwargs.get('options')


        for option in options:
            name = option
            value = option
            if len(option) == 2 and type(option) != str:
                name, value = option

            selected = str(kwargs.get('selected', ''))
            value = str(value)


            self.context.options.append({
                "name" : name,
                "value" : value,
                "selected": "selected" if value == selected else ""
            })


class MultiSelect(UIComponent, pudgy.JinjaComponent, pudgy.BackboneComponent):
    pass

class ControlRow(UIComponent, pudgy.MustacheComponent):
    def __init__(self, name, label, control, hidden=False):
        super(ControlRow, self).__init__(name, label, control)
        self.context.name = name
        self.context.label = label
        self.context.hidden = "hidden" if hidden else ""
        self.context.control = control.__html__



import pudgy
from snorkel.views import ViewBase, get_column_types
from snorkel.components import Table
import os

import numbers

class TableView(ViewBase, pudgy.BackboneComponent, pudgy.SassComponent):
    NAME="table"
    BASE="table"
    DISPLAY_NAME="Table View"
    BASE_DIR=os.path.dirname(__file__)
    SUPPORT_COMPARE_QUERIES=True

    def __render__(self):
        return ""



import os
import pudgy

from snorkel.views import ViewBase

class DistView(ViewBase, pudgy.JSComponent):
    NAME="dist"
    BASE="dist"
    DISPLAY_NAME="Dist View"
    BASE_DIR=os.path.dirname(__file__)
    SUPPORT_COMPARE_QUERIES = True

    def get_controls(self):
        controls = []

        self.add_time_controls(controls)

        self.add_field_selector(controls)
        self.add_go_button(controls)

        return controls



import pudgy

from .components import UIComponent

import flask


# https://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
def sizeof_fmt(num, suffix='B'):
    for unit in ['','Ki','Mi','Gi','Ti','Pi','Ei','Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)

def count_fmt(num, suffix=''):
    for unit in ['','K','M','B','T']:
        if abs(num) < 1000.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1000.0
    return "%.1f%s%s" % (num, 'Z', suffix)

def format_bytes(w):
    return sizeof_fmt(w)

def format_count(w):
    return count_fmt(w)

class StatsBox(UIComponent, pudgy.MustacheComponent):
    def __prepare__(self):
        if self.context.count:
            self.context.object_size = format_bytes(self.context.storageSize / self.context.count)
            self.context.count = format_count(self.context.count)
            self.context.storageSize = format_bytes(self.context.storageSize)

class QueryControls(UIComponent, pudgy.JinjaComponent):
    pass

class QuerySidebar(UIComponent, pudgy.BackboneComponent, pudgy.JinjaComponent, pudgy.SassComponent, pudgy.ServerBridge, pudgy.Pagelet):
    NAMESPACE="views"
    def __prepare__(self):
        from ..views import get_column_types
        self.context.querycontrols = QueryControls(controls=self.context.view.get_controls())
        self.context.viewcontrols = self.context.view.get_view_control()

        self.context.filters = self.context.view.get_filters(self.context.filters)
        self.context.compare_filters = self.context.view.get_filters(self.context.compare_filters)

        self.context.stats = StatsBox(**self.context.metadata)

        fields, types = get_column_types(self.context.metadata)
        self.marshal(fields=fields, types=types,
            supports_compare=self.context.view.SUPPORT_COMPARE_QUERIES)



from __future__ import absolute_import
from .timeline import TimelineView
from .overview import OverviewView
from .forecasting import ForecastView
from .drilldown import DrilldownView
from .weco import WecoView
from .digraph import DigraphView



import flask
from . import backend, rbac

from . import fastjson as json

from .query_spec import QuerySpec
from .views import get_view_by_name
from .util import return_json
import time


import re
COL_RE = re.compile("(.*)\((.*)\)")

def extract_field(col):
    m = COL_RE.match(col)
    if m:
        return m.groups()[1]
    return col

def extract_agg(col):
    m = COL_RE.match(col)
    if m:
        return m.groups()[0].lower()
    return col

def extract_field_aggs(query):
    cols = query.get_fields()

    custom_fields = query.getlist('custom_fields[]')

    col_aggs = {};
    agg = query.get_metric()
    for col in cols:
      col_agg = agg + "(" + col + ")"
      col_aggs[col_agg] = col_agg

    for em in custom_fields:
        col_aggs[em] = em

    return col_aggs.keys()

def run_query(query):
    query = QuerySpec(query)
    table = query.get('table')
    if not rbac.check("query", table):
        return return_json({"error" : "User is not authorized to query this table"})

    d = query.__makedict__()
    bs = backend.SybilBackend()
    ti = bs.get_table_info(table)

    from urllib import unquote
    filters = unquote(query.get("filters", ""))
    if isinstance(filters, (str, unicode)):
        try:
            filters = json.loads(unquote(filters))
            query.set('filters', filters)
        except:
            query.set('filters', { "query" : [], "compare" : []})

    view = query.get('view')
    VwClass = get_view_by_name(view)
    query.set('viewbase', VwClass.BASE)

    start = time.time()
    res = bs.run_query(table, query, ti)

    # need to marshall the series to the format grafana is expecting:
    # it should be an array of objects, each object looks like:
    # { target: name,
    # datapoints: [ [val, timestamp], [val, timestamp], ... ] }

    ret = {}
    groupby = query.get_groupby()
    col_aggs = extract_field_aggs(query)
    if not col_aggs:
        col_aggs = [ "Count" ]

    start = time.time()
    series = {}
    for ts in res:
        row = res[ts]

        for obj in row:
            key = []
            for g in groupby:
                key.append(obj.get(g))
            key = ",".join(key)

            for col in col_aggs:
                kv = "%s .%s" % (key, col)
                agg = extract_agg(col)
                field = extract_field(col)

                d = obj.get(field, None)
                if isinstance(d, dict):
                    if agg[0] == "p":
                        try:
                            val = d["percentiles"][int(agg[1:])]
                        except:
                            val = None
                    else:
                        val = d.get(agg)
                else:
                    val = d

                if not kv in series:
                    series[kv] = {
                        "datapoints" : [],
                        "target" : kv
                    }

                series[kv]["datapoints"].append([val, int(ts) * 1000])

    for k in series:
        series[k]["datapoints"].sort(key=lambda w: w[1])

    return return_json(series.values())



import werkzeug
try:
    import dotmap
except:
    dotmap = None

try:
    import addict
except:
    addict = None

class QuerySpec(object):
    def __init__(self, query):
        # TODO: list all attributes of a query spec up front so others know what to expect

        md = werkzeug.MultiDict()
        for q in query:
            if type(q) == dict:
                md.add(q['name'], q['value'].strip())
            elif type(q) == list or type(q) == tuple:
                md.add(q[0], q[1].strip())
            else:
                md.add(q, query[q])

        self.ismultidict = False
        self.isdotmap = False
        if isinstance(query, werkzeug.MultiDict):
            self.ismultidict = True
        elif addict and isinstance(query, addict.Dict):
            self.isdotmap = True
        elif dotmap and isinstance(query, dotmap.DotMap):
            self.isdotmap = True
        elif isinstance(query, list):
            self.ismultidict = True
        else:
            raise Exception("Unknown entry for query spec")

        self.md = md

        # we will need to put together an exported interface
        self.fields = self.get_fields()
        self.groupby = self.get_groupby()

    def __makedict__(self):
        ret = {


        }
        for f in self.md:
            if f.endswith("[]"):
                if self.ismultidict:
                    ret[f] = self.md.getlist(f)
                else:
                    ret[f] = self.md.get(f)

            else:
                ret[f] = self.md.get(f)

        return ret

    def __json__(self):
        return self.__makedict__()

    def setlist(self, k, v):
        self.md.setlist(k, v)

    def set(self, k, v):
        if k in self.md:
            self.md.pop(k)
        self.md.add(k,v)

    def add(self, k, v):
        self.md.add(k, v)

    def getlist(self, k, d=[]):
        if self.ismultidict:
            return self.md.getlist(k)

        return self.md.get(k) or []



    def get(self, k, d=None):
        return self.md.get(k, d)

    def get_metric(self):
        op = self.md.get('metric')
        if not op:
            op = self.md.get('agg', '')
            op = op.lstrip("$")

        return op

    def get_groupby(self):
        g = self.getlist('groupby[]')
        if not g:
            g = self.getlist('group_by')

        return g

    def get_fields(self):
        g = self.getlist('fields[]')
        if not g:
            g = self.getlist('fieldset')

        return g

    def get_custom_fields(self):
        g = self.getlist('custom_fields[]')
        if not g:
            g = self.getlist('custom_fields')

        return g




import pudgy
import os

from snorkel.components import Selector, ControlRow
from snorkel.views import ViewBase

def make_dict(arr):
    return dict([(w,w) for w in arr])

class TimelineView(ViewBase, pudgy.JSComponent, pudgy.SassComponent):
    NAME="timeline"
    BASE="samples"
    DISPLAY_NAME="Timeline View"
    BASE_DIR=os.path.dirname(__file__)

    def add_timeline_controls(self, controls):
        groups = make_dict(self.context.metadata["columns"]["strs"])
        event_field = Selector(
            name="event_field",
            options=groups,
            selected=self.context.query.get("event_field"))

        controls.append(ControlRow("event_field", "Event Field", event_field))



    def get_controls(self):
        controls = []

        self.add_time_controls(controls)
#        self.add_time_comparison(controls)

        self.add_timeline_controls(controls)

        self.add_groupby_selector(controls)
        self.add_limit_selector(controls)

        self.add_go_button(controls)

        return controls



from __future__ import print_function
# LOAD CONFIGS
import os

CFG_DIR = "config/"

from .default import *

if "ENV" in os.environ:
    print(" * Loading settings from", os.environ["ENV"])
    with open(os.path.join(CFG_DIR, os.environ["ENV"])) as f:
        exec(f.read())



import setuptools

from setuptools import setup

from setuptools.command.build_py import build_py
from setuptools.command.develop import develop
from setuptools.command.install import install
from subprocess import check_call

import os
import shutil

# this builds and installs sybil into the package
class BuildSybilCommand(build_py):
    def run(self):
        gopath = "build/go"
        os.environ["GOPATH"] = os.path.abspath("build/go")
        check_call("/usr/bin/go get -ldflags='-s -w' github.com/logv/sybil", shell=True)

        try:
            os.makedirs("src/backend/bin/")
        except:
            pass
        shutil.copy("build/go/bin/sybil", "src/backend/bin/")

        build_py.run(self)

# TODO: fix permissions for msybil.py and msybil_ingest.py
# from https://stackoverflow.com/questions/20288711/post-install-script-with-python-setuptools
class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        install.run(self)
        for fn in self.get_outputs():
            if fn.find("msybil"):
                # copied from distutils source - make the binaries executable
                mode = ((os.stat(fn).st_mode) | 0o555) & 0o7777
                os.chmod(fn, mode)



try:
    # python2
    execfile('src/version.py')
except NameError as e:
    # python3
    eval('exec(open("./src/version.py").read())')

setup(
    name='snorkel-lite',
    version=__version__,
    author='okay',
    author_email='okayzed+slite@gmail.com',
    include_package_data=True,
    cmdclass={
        'build_py': BuildSybilCommand,
        'install' : PostInstallCommand
    },
    packages=[
        'snorkel',
        'snorkel.backend',
        'snorkel.config',
        'snorkel.views',
        'snorkel.components',
        'snorkel.plugins',
        'snorkel.plugins.snorkel_basic_views',
        'snorkel.plugins.snorkel_advanced_views' ],
    package_dir= { "snorkel" : "src" },
    entry_points = {
        'flask.commands': [
            'add_user=snorkel.cli:add_user',
            'add_superuser=snorkel.cli:add_superuser',
            'get_auth_token=snorkel.cli:get_user_token',
        ],
        'console_scripts': [
            'snorkel.msybil=snorkel.backend.msybil:main',
            'snorkel.msybil_ingest=snorkel.backend.msybil_ingest:main'
        ],

    },

    # url = 'https://github.com/logv/slite',
    # license='MIT',
    description='a data exploration UI',
    long_description=open('README.md').read(),
    install_requires=[ w.strip() for w in open('requirements.txt').readlines()]
    )




from flask import redirect, url_for, request, flash

from flask_admin import Admin, AdminIndexView
from flask_admin.contrib.peewee import ModelView

from flask_security.core import current_user
from .auth import needs_login
from flask_admin import expose

from . import models
# set optional bootswatch theme
class AdminModelView(ModelView):

    def is_accessible(self):
        if not current_user.is_active or not current_user.is_authenticated:
            return False

        if current_user.has_role('superuser'):
            return True

        return False

    def _handle_view(self, name, **kwargs):
        """
        Override builtin _handle_view in order to redirect users when a view is not accessible.
        """

        if not self.is_accessible():
            if current_user.is_authenticated:
                # permission denied
                abort(403)
            else:
                # login
                return redirect(url_for('security.login', next=request.url))

class MyAdminIndexView(AdminIndexView):
    @expose('/')
    @needs_login
    def index(self):
        if current_user.has_role('superuser'):
            return super(MyAdminIndexView,self).index()
        else:
            flash("You must be logged in as a superuser to access /admin")
            return redirect("/")

def install(app):
    app.config['FLASK_ADMIN_SWATCH'] = 'flatly'
    app.config['FLASK_ADMIN_FLUID_LAYOUT'] = True

    admin = Admin(app, name='snorkelite', template_mode='bootstrap3', index_view=MyAdminIndexView())
    admin.add_view(AdminModelView(models.User, name='Users'))
    admin.add_view(AdminModelView(models.UserRoles, name='UserRoles'))
    admin.add_view(AdminModelView(models.Role, name='Roles'))





slides = [
  {
    "src": "/static/images/screens/Screenshot_2016-08-31_07-24-27.png",
    "caption": "",
    "title": "Default view"
  },
  {
    "src": "/static/images/screens/Screenshot_2016-08-31_07-26-49.png",
    "caption": "",
    "title": "Table View"
  },
  {
    "src": "/static/images/screens/gif/table_to_time2.gif",
    "caption": "",
    "title": "Table -> Time"
  },
  {
    "src": "/static/images/screens/gif/time_to_table.gif",
    "caption": "",
    "title": "Time -> Table"
  },
  {
    "src": "/static/images/screens/Screenshot_2016-08-31_07-30-50.png",
    "caption": "",
    "title": "Distribution View (Moments)"
  },
  {
    "src": "/static/images/screens/Screenshot_2016-08-31_11-53-01.png",
    "caption": "",
    "title": "Distribution View (Graphs)"
  },
  {
    "src": "/static/images/screens/gif/table_to_dist2.gif",
    "caption": "",
    "title": "Table -> Grouped Dist View"
  },
  {
    "src": "/static/images/screens/Screenshot_2016-04-30_18-50-20.png",
    "caption": "",
    "title": "Overview"
  },
  {
    "src": "/static/images/screens/Screenshot_2016-04-30_18-51-11.png",
    "caption": "",
    "title": "Overview (cont'd)"
  },
  {
    "src": "/static/images/screens/Screenshot_2016-08-31_07-25-05.png",
    "caption": "",
    "title": "Time Series"
  },
  {
    "src": "/static/images/screens/Screenshot_2016-08-31_07-25-43.png",
    "caption": "",
    "title": "Stacked Time Series"
  },
  {
    "src": "/static/images/screens/gif/forecasting2.gif",
    "caption": "",
    "title": "Time series forecasting"
  },
  {
    "src": "/static/images/screens/Screenshot_2016-08-31_07-32-51.png",
    "caption": "",
    "title": "Anomaly detection"
  },
  {
    "src": "/static/images/screens/Screenshot_2016-04-30_18-51-55.png",
    "caption": "",
    "title": "Bar Charts"
  },
  {
    "src": "/static/images/screens/Screenshot_2016-08-31_07-27-57.png",
    "caption": "",
    "title": "Timelines"
  },
  {
    "src": "/static/images/screens/Screenshot_2016-08-31_07-29-37.png",
    "caption": "",
    "title": "Scatter plots"
  },
  {
    "src": "/static/images/screens/Screenshot_2016-08-31_07-32-23.png",
    "caption": "",
    "title": "Scatter plots (cont'd)"
  },
  {
    "src": "/static/images/screens/Screenshot_2016-04-30_18-52-23.png",
    "caption": "",
    "title": "Scatter plots (cont'd)"
  },
  {
    "src": "/static/images/screens/Screenshot_2016-08-31_07-33-23.png",
    "caption": "",
    "title": "Samples"
  },
#  {
#    "src": "/static/images/screens/1264245_10201979227905947_1254692173_o.jpg",
#    "caption": "",
#    "title": "JSON API"
#  }
]



import pudgy
from snorkel.views import ViewBase, get_column_types
from snorkel.components import Table
import os

class SamplesView(ViewBase, pudgy.JSComponent):
    NAME="samples"
    BASE="samples"
    DISPLAY_NAME="Samples View"
    BASE_DIR=os.path.dirname(__file__)

    def get_controls(self):
        controls = []

        self.add_time_controls(controls)

        self.add_limit_selector(controls)

        self.add_go_button(controls)

        return controls

    def __prepare__(self):
        md = self.context.metadata
        fields, types = get_column_types(md)

        query = self.context.query
        headers = list(fields.keys())
        headers.sort()

        table = []

        for r in self.context.results:
            row = []
            for h in headers:
                row.append(r.get(h, ""))

            table.append(row)

        self.context.table = table
        self.context.headers = headers

    def __render__(self):
        t = Table()
        t.context.update(**self.context)

        r = t.render()

        return r



from __future__ import print_function
import re

METRIC_OPTIONS = [
    "Avg",
    "Sum",
    "Count",
    "p5",
    "p25",
    "p50",
    "p75",
    "p95",
    "Distinct",
]

VIEW_OPTIONS = [ ]

PRESENTERS = []

class DatasetPresenter(object):
    PRESENTERS = []
    @classmethod
    def add_presenter(regex_rule, config):
        PRESENTERS.append((regex_rule, config))

    def __init__(self, *args, **kwargs):
        self.views = []

    def set_views(self, views):
        self.views = views

    def get_views(self):
        view_options = []
        for view in self.views:
            view_options.append(view)

        ret = []
        for view in view_options:
            ret.append((view.get_display_name(), view.get_name()))

        return ret

    def get_metrics(self):
        return METRIC_OPTIONS

# adds a regex and presenter
def RegisterPresenter(rule, presenter):
    re.compile(rule)
    PRESENTERS.append((rule, presenter))


# given a dataset, returns its presenter (using a regex)
def GetPresenter(dataset):
    for r, p in reversed(PRESENTERS):
        if re.search(r, dataset):
            return p


HIDDEN_TABLES = {}
TABLE_NAMES = {}
def SetTableName(realname, name):
    TABLE_NAMES[realname] = name

def SetTableNames(name_lookup):
    for name in name_lookup:
        SetTableName(name, name_lookup[name])

def GetTableName(name):
    if name in TABLE_NAMES:
        return TABLE_NAMES[name]

    return name

def GetRealTable(name):
    for n in TABLE_NAMES:
        if TABLE_NAMES[n] == name:
            return n

    return name

def HideTables(*names):
    for n in names:
        HIDDEN_TABLES[n] = True

def IsTableHidden(name):
    return name in HIDDEN_TABLES

def IsTableVisible(name):
    return not IsTableHidden(name)

# PARAMETERS:
# name=<str> - name to display when showing the table
# hidden=<bool> - whether to hide the table or not
def ConfigureTable(table, **kwargs):
    if "name" in kwargs:
        SetTableName(table, kwargs.get("name"))

    if kwargs.get('hidden', False):
        HideTables(table)



import fnmatch

from . import config
from flask_security import current_user

class PermGroup(object):
    def __init__(self, name, perms, tables, users):
        self.name = name
        self.perms = perms.split(',')
        self.tables = tables.split(',')
        self.users = users.split(',')

    def __repr__(self):
        return '<%s> perms: %s, tables: %s, users: %s' \
            % (self.name, self.perms, self.tables, self.users)

    # % means to use the perms from the group with %NAME
    # $ means to use the users from the group with $NAME
    # ^ means to use the tables from the group with ^NAME
    def add_subs(self, lookup):
        for u in self.users:
            if u[0] == "%":
                name = u[1:]
                self.perms = lookup[name].perms

            if u[0] == "^":
                name = u[1:]
                self.tables = lookup[name].tables

            if u[0] == "$":
                name = u[1:]
                self.users = lookup[name].users
        pass

    def check(self, perm, table, user):
        has_perm = False
        for p in self.perms:
            if fnmatch.fnmatch(perm, p):
                has_perm = True

        if not has_perm:
            return

        has_user = False
        for u in self.users:
            if fnmatch.fnmatch(user, u):
                has_user = True

        if not has_user:
            return

        for t in self.tables:
            if fnmatch.fnmatch(table, t):
                return True

def parse_perms(lines):
    perm_lookups = {}

    for line in lines:
        line = line.strip()
        if not line or line[0] == '#':
            continue

        tokens = line.split(':')
        pg = PermGroup(*tokens)
        perm_lookups[pg.name] = pg

    for p in perm_lookups:
        pg = perm_lookups[p]
        pg.add_subs(perm_lookups)

    return perm_lookups

def read_rbac_file(fname):
    with open(fname) as f:
        lines = f.readlines()

    return parse_perms(lines)

def check(role, table, user=None):
    # we aren't using RBAC check
    if not config.AUTHORIZED_USERS:
        return True

    if not user and current_user.is_anonymous:
        return False

    user = user or current_user.email

    for p in PERMS:
        pg = PERMS[p]
        if pg.check(role, table, user):
            return True

if config.AUTHORIZED_USERS:
    PERMS = read_rbac_file(config.AUTHORIZED_USERS)

if __name__ == "__main__":
    check("admin", "snorkel", "okay")
    check("query", "snorkel", "okay")
    check("admin", "demo", "test")
    check("query", "demo", "test")
